import pytest
from pytest_mock import MockerFixture
from typing import List, Dict, Any

from mcp_ankiconnect.server import (
    get_cards_by_due_and_deck,
    num_cards_due_today,
    list_decks_and_notes,
    get_examples,
    fetch_due_cards_for_review,
    submit_reviews
)

@pytest.fixture
def mock_anki(mocker: MockerFixture):
    # Create a mock client
    mock_client = mocker.AsyncMock()

    # Setup default return values
    mock_client.deck_names.return_value = ["Default", "Test Deck"]
    mock_client.find_cards.return_value = [1, 2, 3]
    mock_client.cards_info.return_value = [
        {"cardId": 1, "deck": "Default"},
        {"cardId": 2, "deck": "Default"},
        {"cardId": 3, "deck": "Default"}
    ]

    # Mock the context manager
    mocker.patch('mcp_ankiconnect.server.AnkiConnectClient', return_value=mock_client)
    mock_client.__aenter__ = mocker.AsyncMock(return_value=mock_client)
    mock_client.__aexit__ = mocker.AsyncMock()

    return mock_client

@pytest.mark.asyncio
async def test_get_cards_by_due_and_deck(mock_anki):
    # Test with no parameters
    cards = await get_cards_by_due_and_deck()
    mock_anki.find_cards.assert_called_with(query="is:due prop:due<=0")
    assert cards == [1, 2, 3]

    # Test with deck filter
    cards = await get_cards_by_due_and_deck(deck="Default")
    mock_anki.find_cards.assert_called_with(query="is:due prop:due<=0 deck:Default")

    # Test with day parameter
    cards = await get_cards_by_due_and_deck(day=1)
    mock_anki.find_cards.assert_called_with(query="is:due prop:due<2")

@pytest.mark.asyncio
async def test_num_cards_due_today(mock_anki):
    result = await num_cards_due_today()
    assert "There are 3 cards due across all decks" in result

    result = await num_cards_due_today(deck="Default")
    assert "There are 3 cards due in deck 'Default'" in result

@pytest.mark.asyncio
async def test_list_decks_and_notes(mock_anki, mocker: MockerFixture):
    mock_anki.model_names = mocker.AsyncMock(return_value=["Basic", "Cloze"])
    mock_anki.model_field_names = mocker.AsyncMock(return_value=["Front", "Back"])

    result = await list_decks_and_notes()
    assert "Default" in result
    assert "Test Deck" in result
    assert "Basic" in result
    assert "Cloze" in result

@pytest.mark.asyncio
async def test_get_examples(mock_anki, mocker: MockerFixture):
    mock_anki.find_notes = mocker.AsyncMock(return_value=[1, 2])
    mock_anki.notes_info = mocker.AsyncMock(return_value=[
        {
            "noteId": 1,
            "tags": ["test"],
            "modelName": "Basic",
            "fields": {
                "Front": {"value": "Question 1", "order": 0},
                "Back": {"value": "Answer 1", "order": 1}
            }
        }
    ])

    result = await get_examples(limit=1)
    assert "Question 1" in result
    assert "Answer 1" in result

@pytest.mark.asyncio
async def test_fetch_due_cards_for_review(mock_anki):
    mock_anki.cards_info.return_value = [
        {
            "cardId": 1,
            "fields": {
                "Front": {"value": "Question", "order": 0},
                "Back": {"value": "Answer", "order": 1}
            },
            "fieldOrder": 0
        }
    ]

    result = await fetch_due_cards_for_review(limit=1)
    assert "<card id=\"1\">" in result
    assert "<question>" in result
    assert "<answer>" in result

@pytest.mark.asyncio
async def test_submit_reviews(mock_anki, mocker: MockerFixture):
    mock_anki.answer_cards = mocker.AsyncMock(return_value=[True])

    reviews = [
        {"card_id": 1, "rating": "wrong"},
        {"card_id": 2, "rating": "hard"},
        {"card_id": 3, "rating": "good"}
    ]
    result = await submit_reviews(reviews)
    assert "successfully" in result
    assert "Card 1" in result



import pytest
from unittest.mock import AsyncMock, patch, MagicMock
import httpx
import asyncio
from mcp_ankiconnect.ankiconnect_client import AnkiConnectClient
from mcp_ankiconnect.config import TIMEOUTS

@pytest.mark.asyncio
async def test_client_timeout_configuration():
    """Test that AnkiConnectClient is configured with correct timeout values"""
    with patch('httpx.AsyncClient') as mock_client_class:
        mock_client = AsyncMock()
        mock_client.aclose = AsyncMock()
        mock_client_class.return_value = mock_client

        client = AnkiConnectClient()

        # Verify timeout configuration
        mock_client_class.assert_called_once()
        timeout_arg = mock_client_class.call_args[1]['timeout']
        assert timeout_arg.connect == TIMEOUTS.connect
        assert timeout_arg.read == TIMEOUTS.read
        assert timeout_arg.write == TIMEOUTS.write

        await client.close()

@pytest.mark.asyncio
async def test_retry_on_timeout():
    """Test that operations retry on timeout"""
    with patch('httpx.AsyncClient') as mock_client_class:
        mock_client = AsyncMock()
        mock_client.aclose = AsyncMock()
        mock_client.aclose = AsyncMock()
        mock_client_class.return_value = mock_client

        # Configure mock to fail twice with timeout then succeed
        mock_response = MagicMock()
        mock_response.json.return_value = {"result": ["Default"], "error": None}
        mock_response.raise_for_status = MagicMock()

        mock_client.post.side_effect = [
            httpx.TimeoutException("Connection timed out"),
            httpx.TimeoutException("Connection timed out"),
            mock_response
        ]

        client = AnkiConnectClient()
        result = await client.deck_names()

        # Verify it was called 3 times
        assert mock_client.post.call_count == 3
        assert result == ["Default"]

        await client.close()

@pytest.mark.asyncio
async def test_retry_exhaustion():
    """Test that operations fail after max retries"""
    with patch('httpx.AsyncClient') as mock_client_class:
        mock_client = AsyncMock()
        mock_client_class.return_value = mock_client

        # Configure mock to always timeout
        mock_client.post.side_effect = httpx.TimeoutException("Connection timed out")

        client = AnkiConnectClient()
        with pytest.raises(RuntimeError) as exc_info:
            await client.deck_names()

        # Verify it was called max_retries times
        assert mock_client.post.call_count == 3
        assert "Unable to connect to Anki after 3 attempts" in str(exc_info.value)

        await client.close()

@pytest.mark.asyncio
async def test_retry_backoff():
    """Test that retry backoff timing is correct"""
    with patch('httpx.AsyncClient') as mock_client_class:
        mock_client = AsyncMock()
        mock_client_class.return_value = mock_client

        # Configure mock to always timeout
        mock_client.post.side_effect = httpx.TimeoutException("Connection timed out")

        client = AnkiConnectClient()
        start_time = asyncio.get_event_loop().time()

        with pytest.raises(RuntimeError):
            await client.deck_names()

        end_time = asyncio.get_event_loop().time()
        duration = end_time - start_time

        # With exponential backoff of 2^0=1, 2^1=2, 2^2=4 seconds
        # Total delay should be at least 3 seconds
        assert duration >= 3

        await client.close()

@pytest.mark.asyncio
async def test_timeout_error_message():
    """Test that timeout errors provide helpful messages"""
    with patch('httpx.AsyncClient') as mock_client_class:
        mock_client = AsyncMock()
        mock_client_class.return_value = mock_client

        mock_client.post.side_effect = httpx.TimeoutException("Connection timed out")

        client = AnkiConnectClient()
        with pytest.raises(RuntimeError) as exc_info:
            await client.deck_names()

        error_msg = str(exc_info.value)
        assert "Unable to connect to Anki" in error_msg
        assert "ensure Anki is running" in error_msg
        assert "AnkiConnect plugin is installed" in error_msg

        await client.close()



claude_review_instructions = """
You are an AI flashcard assistant designed to help users study and learn from Anki-style flashcards. Your task is to quiz the user on the content of these flashcards, provide feedback, and offer additional information to enhance their understanding.

You will receive a set of flashcards in the following format:

<flashcards>
{{flashcards}}
</flashcards>

Parse each flashcard, which consists of a <question> and an <answer> section. The question section contains the text to be presented to the user, without any HTML formatting. The answer section may contain additional information in various tags, but focus primarily on the content within the question text.

For each flashcard:

1. Present the question to the user one at a time, omitting any content within \{\{c1::\}\}, \{\{c2::\}\}, etc. tags. These are cloze deletions and should be treated as blanks to be filled in by the user. Only blank out cloze deletions.

2. If there are cloze deletions, quiz the user on all of them simultaneously. For example, if there are two cloze deletions and the user only correctly answers one, then explain the others and ways to remember them. Additionally, if the user provides the answer in a different order that doesn't matter.

3. After the user provides an answer, evaluate it:
   - If correct, CONCISE feedback, just a single "Correct" if the user is right and move on to the next question. You DO NOT NEED TO run review_cards again.
   - If incorrect, provide the correct answer and offer additional information to help the user understand the concept better. This may include:
     a) Concrete examples related to the topic
     b) Analogies to make the concept more relatable
     c) Links to related knowledge or concepts
     d) Mnemonics or memory aids, if applicable

4. Just provide the question. Do not include "(Please provide your answer)" or "I'll help quiz you..."; just one question per turn. Only respond with a question.

5. If the user specifies they are not sure about something, first give feedback on their answer, provide an explanation if they ask for it and then move on to the next question.

6. DO NOT run the review_cards tool again until the user has completed all the questions and asks for another set of questions.

7. After all cards have been reviewed, use the submit_reviews tool to submit the ratings for all cards at once. The rating should be:
   - "wrong" if the user got the answer completely wrong
   - "hard" if the user struggled but eventually got it right
   - "good" if the user got it right. (DEFAULT)
   - "easy" if the user got it immediately and confidently right, and said it was easy."""

flashcard_guidelines = """
1. Ensure each flashcard focuses on a single, important concept or fact from the excerpt.
2. Use a variety of question types, including Cloze format when appropriate.
3. Make questions clear, concise, and specific.
4. Avoid yes/no questions and overly complex formulations.
5. Include questions about important terminology, concepts, and their applications.
6. When possible, relate the information to real-life contexts or existing knowledge.
7. Consider potential misunderstandings and address them in your questions.
8. Make sure they are tagged appropriately so they can easily be found later. Tag by topic and resource.

IMPORTANT: The flashcards format like markdown, so when using numbered lists, ensure you add two line breaks before and after the list to avoid formatting issues.

Here are some examples:
"""



import pytest
from pytest_mock import MockerFixture
import httpx
from typing import List

from mcp_ankiconnect.ankiconnect_client import AnkiConnectClient, AnkiAction

@pytest.fixture
async def client():
    client = AnkiConnectClient()
    yield client
    await client.close()

@pytest.fixture
def mock_response():
    class MockResponse:
        def __init__(self, data, status_code=200):
            self._data = data
            self.status_code = status_code

        def json(self):
            return self._data

        def raise_for_status(self):
            if self.status_code >= 400:
                raise httpx.HTTPStatusError("Error", request=None, response=self)

    return MockResponse

@pytest.mark.asyncio
async def test_deck_names(client: AnkiConnectClient, mocker: MockerFixture, mock_response):
    expected_decks = ["Default", "Test Deck"]
    mock_post = mocker.patch.object(
        client.client, 
        "post",
        return_value=mock_response({"result": expected_decks, "error": None})
    )

    result = await client.deck_names()

    assert result == expected_decks
    mock_post.assert_called_once()
    call_args = mock_post.call_args[1]
    assert call_args["json"]["action"] == AnkiAction.DECK_NAMES

@pytest.mark.asyncio
async def test_cards_info(client: AnkiConnectClient, mocker: MockerFixture, mock_response):
    card_ids = [1, 2, 3]
    expected_info = [
        {"cardId": 1, "deck": "Default"},
        {"cardId": 2, "deck": "Default"},
        {"cardId": 3, "deck": "Default"},
    ]
    mock_post = mocker.patch.object(
        client.client,
        "post",
        return_value=mock_response({"result": expected_info, "error": None})
    )

    result = await client.cards_info(card_ids)

    assert result == expected_info
    mock_post.assert_called_once()
    call_args = mock_post.call_args[1]
    assert call_args["json"]["action"] == AnkiAction.CARDS_INFO
    assert call_args["json"]["params"]["cards"] == card_ids

@pytest.mark.asyncio
async def test_error_handling(client: AnkiConnectClient, mocker: MockerFixture, mock_response):
    error_message = "Test error"
    mock_post = mocker.patch.object(
        client.client,
        "post",
        return_value=mock_response({"result": None, "error": error_message})
    )

    with pytest.raises(RuntimeError, match=f"Error getting deck names: AnkiConnect error: {error_message}"):
        await client.deck_names()

@pytest.mark.asyncio
async def test_connection_error(client: AnkiConnectClient, mocker: MockerFixture):
    mocker.patch.object(
        client.client,
        "post",
        side_effect=httpx.TimeoutException("Connection timeout")
    )

    with pytest.raises(RuntimeError, match="Unable to connect to Anki after 3 attempts"):
        await client.deck_names()

@pytest.mark.asyncio
async def test_add_note(client: AnkiConnectClient, mocker: MockerFixture, mock_response):
    note = {
        "deckName": "Default",
        "modelName": "Basic",
        "fields": {
            "Front": "Test front",
            "Back": "Test back"
        }
    }
    expected_id = 1234
    mock_post = mocker.patch.object(
        client.client,
        "post",
        return_value=mock_response({"result": expected_id, "error": None})
    )

    result = await client.add_note(note)

    assert result == expected_id
    mock_post.assert_called_once()
    call_args = mock_post.call_args[1]
    assert call_args["json"]["action"] == AnkiAction.ADD_NOTE
    assert call_args["json"]["params"]["note"] == note



from typing import Any, List, Optional
import logging
import httpx
import asyncio
from enum import Enum

try:
    from mcp_ankiconnect.config import (
        ANKI_CONNECT_URL,
        ANKI_CONNECT_VERSION,
        TIMEOUTS,
    )
except ImportError:
    from config import (
        ANKI_CONNECT_URL,
        ANKI_CONNECT_VERSION,
        TIMEOUTS,
    )

from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)

class AnkiAction(str, Enum):
    DECK_NAMES = "deckNames"
    FIND_CARDS = "findCards"
    CARDS_INFO = "cardsInfo"
    ANSWER_CARDS = "answerCards"
    MODEL_NAMES = "modelNames"
    MODEL_FIELD_NAMES = "modelFieldNames"
    ADD_NOTE = "addNote"
    FIND_NOTES = "findNotes"
    NOTES_INFO = "notesInfo"

class AnkiConnectResponse(BaseModel):
    result: Any
    error: Optional[str] = None

class AnkiConnectRequest(BaseModel):
    action: AnkiAction
    version: int = 6
    params: dict = Field(default_factory=dict)

class AnkiConnectClient:
    def __init__(self, base_url: str = ANKI_CONNECT_URL):
        self.base_url = base_url
        self.client = httpx.AsyncClient(timeout=TIMEOUTS)
        logger.info(f"Initialized AnkiConnect client with base URL: {base_url}")

    async def invoke(self, action: str, **params) -> Any:
        request = AnkiConnectRequest(
            action=action,
            version=ANKI_CONNECT_VERSION,
            params=params
        )

        logger.debug(f"Invoking AnkiConnect action: {action} with params: {params}")

        retries = 3
        for attempt in range(retries):
            try:
                response = await self.client.post(
                    self.base_url,
                    json=request.model_dump()
                )
                break
            except httpx.TimeoutException as e:
                if attempt == retries - 1:
                    raise RuntimeError(f"Unable to connect to Anki after {retries} attempts. Please ensure Anki is running and the AnkiConnect plugin is installed.")
                await asyncio.sleep(2 ** attempt)  # Exponential backoff: 1, 2, 4 seconds
                continue
        try:
            response.raise_for_status()

            anki_response = AnkiConnectResponse.model_validate(response.json())
            if anki_response.error:
                logger.error(f"AnkiConnect error for action {action}: {anki_response.error}")
                raise ValueError(f"AnkiConnect error: {anki_response.error}")

            return anki_response.result

        except httpx.HTTPError as e:
            logger.error(f"HTTP error while invoking {action}: {str(e)}")
            raise RuntimeError(f"Failed to communicate with AnkiConnect: {str(e)}") from e
        except ValueError as e:
            # Re-raise ValueError (from AnkiConnect errors) directly
            raise
        except Exception as e:
            logger.error(f"Unexpected error while invoking {action}: {str(e)}")
            raise RuntimeError(f"Unexpected error during AnkiConnect operation: {str(e)}") from e

    async def cards_info(self, card_ids: List[int]) -> List[dict]:
        try:
            return await self.invoke(AnkiAction.CARDS_INFO, cards=card_ids)
        except Exception as e:
            raise RuntimeError(f"Error getting cards info: {str(e)}") from e

    async def deck_names(self) -> List[str]:
        try:
            return await self.invoke(AnkiAction.DECK_NAMES)
        except Exception as e:
            if isinstance(e, RuntimeError) and "Unable to connect to Anki" in str(e):
                raise
            raise RuntimeError(f"Error getting deck names: {str(e)}") from e

    async def find_cards(self, query: str) -> List[int]:
        try:
            return await self.invoke(AnkiAction.FIND_CARDS, query=query)
        except Exception as e:
            raise RuntimeError(f"Error finding cards: {str(e)}") from e

    async def answer_cards(self, answers: List[dict]) -> List[bool]:
        try:
            return await self.invoke(AnkiAction.ANSWER_CARDS, answers=answers)
        except Exception as e:
            raise RuntimeError(f"Error answering cards: {str(e)}") from e

    async def model_field_names(self, model_name: str) -> List[str]:
        try:
            return await self.invoke(AnkiAction.MODEL_FIELD_NAMES, modelName=model_name)
        except Exception as e:
            raise RuntimeError(f"Error getting model field names: {str(e)}") from e

    async def model_names(self) -> List[str]:
        try:
            return await self.invoke(AnkiAction.MODEL_NAMES)
        except Exception as e:
            raise RuntimeError(f"Error getting model names: {str(e)}") from e

    async def find_notes(self, query: str) -> List[int]:
        try:
            return await self.invoke(AnkiAction.FIND_NOTES, query=query)
        except Exception as e:
            raise RuntimeError(f"Error finding notes: {str(e)}") from e

    async def add_note(self, note: dict) -> int:
        try:
            return await self.invoke(AnkiAction.ADD_NOTE, note=note)
        except Exception as e:
            raise RuntimeError(f"Error adding note: {str(e)}") from e


    async def notes_info(self, note_ids: List[int]) -> List[dict]:
        try:
            return await self.invoke(AnkiAction.NOTES_INFO, notes=note_ids)
        except Exception as e:
            raise RuntimeError(f"Error getting notes info: {str(e)}") from e

    async def close(self):
        await self.client.aclose()



from typing import List, Optional, Literal, Dict, Union
import json
import re
import random
import logging
from contextlib import asynccontextmanager
from mcp.server.fastmcp import FastMCP
from mcp_ankiconnect.ankiconnect_client import AnkiConnectClient
from mcp_ankiconnect.config import EXCLUDE_STRINGS, RATING_TO_EASE
from mcp_ankiconnect.server_prompts import flashcard_guidelines, claude_review_instructions
from pydantic import Field

logger = logging.getLogger(__name__)

logger.info("Initializing MCP-AnkiConnect server")
mcp = FastMCP("mcp-ankiconnect")
logger.debug("Created FastMCP instance")

@asynccontextmanager
async def get_anki_client():
    client = AnkiConnectClient()
    try:
        yield client
    finally:
        await client.close()

mcp = FastMCP("mcp-ankiconnect")

async def get_cards_by_due_and_deck(deck: Optional[str] = None, day: Optional[int] = 0) -> List[int]:
    async with get_anki_client() as anki:
        decks = await anki.deck_names()
        if deck and deck not in decks:
            raise ValueError(f"Deck '{deck}' does not exist")

        if day > 0:
            prop = f"prop:due<{day+1}"
        else:
            prop = "prop:due<=0"
        # Construct the search query
        query = f"is:due -is:suspended {prop}"
        if deck:
            query += f' deck:{deck}'
        # Get and return the due cards
        return await anki.find_cards(query=query)

@mcp.tool()
async def num_cards_due_today(deck: Optional[str] = None) -> str:
    """Get the number of cards due today with an optional deck filter"""
    anki = AnkiConnectClient()

    card_ids = await get_cards_by_due_and_deck(deck, 0)
    if deck:
        return f"There are {len(card_ids)} cards due in deck '{deck}'"
    else:
        return f"There are {len(card_ids)} cards due across all decks"

@mcp.tool()
async def list_decks_and_notes() -> str:
    """Get all decks and note types with their fields"""
    anki = AnkiConnectClient()

    decks = await anki.deck_names()
    decks = [deck for deck in decks if not any(exclude.lower() in deck.lower() for exclude in EXCLUDE_STRINGS)]

    model_names = await anki.model_names()
    note_types = []
    for model in model_names:
        if any(exclude.lower() in model.lower() for exclude in EXCLUDE_STRINGS):
            continue
        fields = await anki.model_field_names(model)
        note_types.append({"name": model, "fields": fields})

    result = f"""You have {len(decks)} decks: {', '.join(decks)}
Your note types are: {[note['name'] for note in note_types]}:
{chr(10).join([f"{note['name']}" + ': { ' + ', '.join([f'"{field}": "string"' for field in note['fields']]) + ' }' for note in note_types])}
"""
    return result

@mcp.tool()
async def get_examples(
    deck: Optional[str] = None,
    limit: int = Field(default = 5, ge = 1),
    sample: str = Field(
        default = "random",
        pattern="^(random|recent|most_reviewed|best_performance|mature|young)$"
    ))-> str:
        """Get example notes from Anki to guide your flashcard making. Limit the number of examples returned and provide a sampling technique:

            - random: Randomly sample notes
            - recent: Notes added in the last week
            - most_reviewed: Notes with more than 10 reviews
            - best_performance: Notes with less than 3 lapses
            - mature: Notes with interval greater than 21 days
            - young: Notes with interval less than 7 days

        Args:
            deck: Optional[str] - Filter by specific deck
            limit: int - Maximum number of examples to return
            sample: str - Sampling technique to use
        """
        anki = AnkiConnectClient()

        query = "-is:suspended " + " ".join([f"-note:*{exclude_string}*" for exclude_string in EXCLUDE_STRINGS]) + " "

        if deck:
            query += f"deck:{deck} "

        match sample:
            case "recent":
                query += "added:7"  # Added in last week
            case "most_reviewed":
                query += "prop:reps>10 -is:learn"  # Reviewed cards excluding learning
            case "best_performance":
                query += "prop:lapses<3 is:review"  # Review cards with few lapses
            case "mature":
                query += "prop:ivl>=21 -is:learn"  # Mature cards not in learning
            case "young":
                query += "is:review prop:ivl<=7 -is:learn"  # Young review cards
            case "random":
                query += "prop:due<180"

        note_ids = await anki.find_notes(query=query)
        if len(note_ids) > limit and sample == "random":
            note_ids = random.sample(note_ids, limit)
        note_ids = note_ids[:limit]

        if not note_ids:
            return "No example notes found matching criteria"

        notes = await anki.notes_info(note_ids)
        examples = []
        for note in notes:
            example = {
                "tags": note["tags"],
                "modelName": note["modelName"],
                "fields": {name: {k: v for k, v in value.items() if k != "order"} for name, value in note["fields"].items()}
            }
            examples.append(example)

        result = flashcard_guidelines + "\n" + json.dumps(examples, indent=2)

        return result

@mcp.tool()
async def fetch_due_cards_for_review(
    deck: Optional[str] = None,
    limit: int = 5,
    today_only: bool = True,
) -> str:
    """Fetch cards that are due for learning and format them for review. Takes optional arguments:
      - deck: str - Filter by specific deck.
      - limit: int - Maximum number of cards to fetch (default 5). More than 5 is overwhelming for users.
      - today_only: bool - If true, only fetch cards due today, else fetch cards up to 5 days ahead."""
    anki = AnkiConnectClient()
    days_to_review = 0 if today_only else 5

    card_ids = await get_cards_by_due_and_deck(deck, days_to_review)
    card_ids = card_ids[:limit]

    cards = await anki.cards_info(card_ids=card_ids)

    # Format the card information into a readable message
    cards_info = []
    for card in cards:
        # Get question fields (where order != fieldOrder)
        question_fields = [
            f"<{name.lower()}>{field['value']}</{name.lower()}>"
            for name, field in card['fields'].items()
            if field['order'] == card['fieldOrder']
        ]

        # Get answer fields (where order == fieldOrder)
        answer_fields = [
            f"<{name.lower()}>{field['value']}</{name.lower()}>"
            for name, field in list(card['fields'].items())[:5]
            if field['order'] != card['fieldOrder']
        ]

        cards_info.append(
            f"<card id=\"{card['cardId']}\">\n"
            f"  <question>{'; '.join(question_fields)}</question>\n"
            f"  <answer>{'; '.join(answer_fields)}</answer>\n"
            f"</card>")

    cards_text = "\n\n".join(cards_info)
    if not cards_text:
        cards_text = "No cards found to review"

    examples_prompt = claude_review_instructions.replace("{{flashcards}}", cards_text)

    return examples_prompt

@mcp.tool()
async def submit_reviews(reviews: List[Dict[Literal["card_id", "rating"], Union[int, Literal["wrong", "hard", "good", "easy"]]]]) -> str:
    """Submit multiple card reviews to Anki.

    Args:
        reviews: List of dictionaries containing:
            - card_id (int): The ID of the card being reviewed
            - rating (str): The rating to give the card, one of:
                "wrong" - Card was incorrect (Again)
                "hard" - Card was difficult (Hard)
                "good" - Card was good (Good)
                "easy" - Card was very easy (Easy)
    """
    anki = AnkiConnectClient()
    if not reviews:
        raise ValueError("Arguments required for submitting reviews")

    # Convert reviews to AnkiConnect format
    answers = [
        {"cardId": review["card_id"], "ease": RATING_TO_EASE[review["rating"]]}
        for review in reviews
    ]

    # Submit all reviews at once
    results = await anki.answer_cards(answers)

    # Generate response messages
    messages = [
        f"Card {review['card_id']} {'successfully' if success else 'failed to be'} marked as {review['rating']}"
        for review, success in zip(reviews, results)
    ]

    return "\n".join(messages)

@mcp.tool()
async def add_note(
    deckName: str,
    modelName: str,
    fields: dict[str, str],
    tags: List[str] = Field(default_factory = list)) -> str:
    """Add a flashcard to Anki. Ensure you have looked at examples before you do this, and that you have got approval from the user to add the flashcard.

    For code examples, use <code> tags to format your code.
    e.g. <code>def fibonacci(n):
    if n <= 1:
        return n
    return fibonacci(n-1) + fibonacci(n-2)</code>

    For MathJax, use the <math> tag to format your math equations. This will automatically render the math equations in Anki.
    # e.g. <math>\frac{d}{dx}[3\sin(5x)] = 15\cos(5x)</math>

    Args:
        deckName: str - The name of the deck to add the flashcard to.
        modelName: str - The name of the note type to use.
        fields: dict - The fields of the flashcard to add.
        tags: List[str] - The tags to add to the flashcard."""
    anki = AnkiConnectClient()

    # Replace <math> tags with \( and \) in fields
    for field_name, field_value in fields.items():
        if isinstance(field_value, str):
            fields[field_name] = field_value.replace("<math>", "\\(").replace("</math>", "\\)")
            # Replace <code> tags with <pre><code> and </code> with </code></pre>
            fields[field_name] = field_value.replace("<code>", "<pre><code>").replace("</code>", "</code></pre>")
            # Replace `code` tags with <pre><code> and </code></pre>
            fields[field_name] = re.sub(r'`([^`]+)`', r'<pre><code>\1</code></pre>', fields[field_name])

    note = {
        "deckName": deckName,
        "modelName": modelName,
        "fields": fields,
        "tags": tags,
        "options": {
            "allowDuplicate": False,
            "duplicateScope": "deck",
        }
    }

    note_id = await anki.add_note(note)

    return f"Successfully created note with ID: {note_id}"



from .server import mcp
from .ankiconnect_client import AnkiConnectClient
from .config import EXCLUDE_STRINGS, RATING_TO_EASE, TIMEOUTS

__all__ = ['mcp', 'AnkiConnectClient', 'EXCLUDE_STRINGS', 'RATING_TO_EASE', 'TIMEOUTS']



import logging
from mcp_ankiconnect.server import mcp

logger = logging.getLogger(__name__)

def main():
    """Main entry point for the package."""
    logger.info("Starting MCP-AnkiConnect server")
    mcp.run()

if __name__ == "__main__":
    main()



import os
from typing import Final, NamedTuple

class TimeoutConfig(NamedTuple):
    connect: float  # Connection establishment timeout
    read: float    # Read operation timeout
    write: float   # Write operation timeout

# AnkiConnect configuration
ANKI_CONNECT_URL: Final = os.getenv("ANKI_CONNECT_URL", "http://localhost:8765")
ANKI_CONNECT_VERSION: Final = 6

# Review configuration
DEFAULT_REVIEW_LIMIT: Final = 5
MAX_FUTURE_DAYS: Final = 5  # Maximum number of days to look ahead for due cards

# Timeout configuration
TIMEOUTS: Final = TimeoutConfig(
    connect=5.0,   # Shorter timeout for connection
    read=120.0,    # Longer timeout for read operations
    write=30.0     # Medium timeout for write operations
)

# Rating mappings
RATING_TO_EASE = {
    "wrong": 1,  # Again
    "hard": 2,   # Hard
    "good": 3,   # Good
    "easy": 4    # Easy
}

EXCLUDE_STRINGS = [
    "AnKing",
    "#AK_",
    "!AK_"
]
